.navbar {
    margin-bottom: 0px;
}

.page-header {
    margin-top: 22px;
    position: sticky;
    top: 0;
    background-color: #fff;
    margin-bottom: 3px;
    border-bottom: 0;
    box-shadow: 0px 0px 8px 8px #fff;
}
.page-header h1 {
    margin-bottom: 0;
}

.categoryHeader {
    font-size: 24px;
    color: #777;
    margin-bottom: 1.8em;
}

a[name] {
    position: relative;
    bottom: 62px;
}

ul {
    margin-top: 10px;
    padding-left: 10px;
}

li {
    padding-top: 5px;
    padding-bottom: 5px;
}

li a {
    text-decoration: none;
}

ul {
    margin-left: 10px;
}

ul ul {
    border-left-color: #e1f5fe;
    border-left-width: 1px;
    border-left-style: solid;
}

ul ul ul {
    border-left-color: #b3e5fc;
}

ul ul ul ul {
    border-left-color: #81d4fa;
}

ul ul ul ul ul {
    border-left-color: #4fc3f7;
}

ul ul ul ul ul ul {
    border-left-color: #29b6f6;
}

ul ul ul ul ul ul ul {
    border-left-color: #03a9f4;
}

ul ul ul ul ul ul ul {
    border-left-color: #039be5;
}

.pre {
    white-space: pre;
    font-family: Menlo,Monaco,Consolas,"Courier New",monospace;
}

.undocumented {
    font-style: italic;
    color: #9e9e9e;
}

.functionBody p {
    margin-top: 6px;
    margin-bottom: 6px;
}

#splitTables > p {
    margin-bottom: 5px;
}

#splitTables > table, .fieldTable {
    margin-bottom: 20px;
    width: 100%;
    border: 0;
}

#splitTables > table {
    border: 1px solid #eee;
}

#splitTables > table tr {
    border-bottom-color: #eee;
    border-bottom-width: 1px;
    border-bottom-style: solid;
}

#splitTables > table tr td, .fieldTable tr td {
    padding: 5px;
}

#splitTables > table tr td {
    border-left-color: #eee;
    border-left-width: 1px;
    border-left-style: solid;
}

#splitTables > table tr td:nth-child(1), .fieldTable tr td:nth-child(1) {
    border-left: none;
    width: 150px;
}

#splitTables > table tr td:nth-child(2), .fieldTable tr td.fieldArg {
    font-family: Menlo,Monaco,Consolas,"Courier New",monospace;
    width: 200px;
}

tr.package {
    background-color: #fff3e0;
}

tr.module {
    background-color: #fff8e1;
}

tr.class, tr.classvariable, tr.baseclassvariable {
    background-color: #fffde7;
}

tr.instancevariable, tr.baseinstancevariable, tr.variable, tr.attribute, tr.property {
    background-color: #f3e5f5;
}

tr.interface {
    background-color: #fbe9e7;
}

tr.method, tr.function, tr.basemethod, tr.baseclassmethod, tr.classmethod {
    background-color: #f1f8e9;
}

tr.private {
    background-color: #f1f1f1;
}

.fieldTable {
    margin-top: 10px;
}

.fieldName {
    font-weight: bold;
}


#childList > div {
    margin: 10px;
    padding: 10px;
    padding-bottom: 5px;
}

.functionBody {
    margin-left: 15px;
}

.functionBody > #part {
    font-style: italic;
}

.functionBody > #part > a {
    text-decoration: none;
}

.functionBody .interfaceinfo {
    font-style: italic;
    margin-bottom: 3px;
}

.functionBody > .undocumented {

    margin-top: 6px;
    margin-bottom: 6px;
}

/*
- <h1><code></code></h1> : This is for the main top navagation bar

- <a><code></code><a> for links in docstrings texts
This is inline docstring content marked up as code, for example L{foo} in epytext
or `bar` in restructuredtext

- <a class="code"> links that are present in summary tables
*/
a > code, a.code {
    color:#c7254e;
    background-color:#f9f2f4;
}
.page-header > h1 > code {
    color: #971c3a;
}

/*
This defines  the "code" class, it's black on light gray,

It also overwrite the default values inherited from bootstrap min
*/
code, .code {
    font-family: Menlo,Monaco,Consolas,"Courier New",monospace;
    padding:2px 4px;
    font-size:90%;
    color: #222222;
    background-color: #f4f4f4;
    border-radius:4px
}

code > .code {
    padding: 0px;
}



a.functionSourceLink {
    font-weight: normal;
}


#childList > div {
    border-left-color: #03a9f4;
    border-left-width: 1px;
    border-left-style: solid;
    background: #fafafa;
}

/* Function definitions `def func():` and `var =` */

#childList > div .functionHeader {
    font-family: monospace;
    font-weight: bold;
    color: #222222;

}

.moduleDocstring {
    margin: 20px;
}

#partOf {
    margin-top: -13px;
    margin-bottom: 19px;
}

.fromInitPy {
    font-style: italic;
}

pre {
    padding-left: 0px;
}

/* Private stuff */

body.private-hidden #splitTables .private,
body.private-hidden #childList .private,
body.private-hidden #summaryTree .private {
    display: none;
}

#showPrivate {
    position: fixed;
    bottom: 0;
    left: 0;
    padding: 10px;
    opacity: 0.7;
}

#showPrivate:hover {
    opacity: 1;
}

#current-docs-container {
    font-style: italic;
    padding-top: 11px;
}

/* Deprecation stuff */

.deprecationNotice {
    margin: 10px;
}

/* Syntax highlighting for source code */

.py-string {
    color: #337ab7;
}
.py-comment {
    color: #309078;
    font-style: italic;
}
.py-keyword {
    font-weight: bold;
}
.py-defname {
    color: #a947b8;
    font-weight: bold;
}
.py-builtin {
    color: #fc7844;
    font-weight: bold;
}

/* Doctest */

pre.py-doctest {
    padding: .5em;
}
.py-prompt, .py-more {
    color: #a8a8a8;
}
.py-output {
    color: #c7254e;
}
